/*
 *  Converts a 8bit bitmap file into an ASM
 */
#include <iostream>
#include <vector>
#include <string>
#include <stack>
#include "bmp.hpp"
#include "act.hpp"
#include "parse.hpp"

void parse_opt(char *s);
FILE *cout;
FILE *hout;

// #define DEBUG -1
#define	__USAGE__	\
"Usage: bmp2spr [options] input \n"

static bool debug_print = false;

Bitmap bmp;
void get_basename(std::string &str, const char *c);

//=============================================================================
std::stack<int> stk;

void push(int n)
{
	stk.push(n);
}

int pop()
{
	int t = stk.top();
	stk.pop();
	return t;
}

int size()
{
	return stk.size();
}

void clear()
{
	stk = std::stack<int>();
}


//=============================================================================
struct CStock
{
	struct Info {
		int x, y;
		int x_size, y_size;
		int x_rel, y_rel;
		int hvflip;
	};
	struct Stock {
		std::string obj_name;
		std::string bmp_name;
		std::vector<Info> info;
	};

	std::vector<Stock> stockArray;

	std::string now_bmp;
	std::string now_obj;
	Info tmp_info;
	Info tmp_info_old;
	Stock tmp_stock;

	CStock() {
		now_bmp = "";
		now_obj = "";
		tmp_info.x = -1;
		tmp_info.y = -1;
		tmp_info.x_size = 16;
		tmp_info.y_size = 16;
		tmp_info.x_rel =  0;
		tmp_info.y_rel =  0;
		tmp_info.hvflip = 0;
	}
	~CStock() {}

	void 
	initTemp(void)
	{
		now_obj = "";
		tmp_info.x = -1;
		tmp_info.y = -1;
		tmp_info.x_size = -1;
		tmp_info.y_size = -1;
		tmp_info.x_rel = 0;
		tmp_info.y_rel = 0;
		tmp_info.hvflip = 0;
	}

	void print(Bitmap *b) {
		std::vector<Stock>::iterator citr = stockArray.begin();
		std::vector<Stock>::iterator cend = stockArray.end();
		for ( ; citr != cend; ++citr) {

			std::vector<Info>::iterator iitr = citr->info.begin();
			std::vector<Info>::iterator iend = citr->info.end();

			if (!b->isLoaded(citr->bmp_name)) printf("* ");
			printf("[%s] ",citr->bmp_name.c_str());
			printf("%s : ",citr->obj_name.c_str());
			for ( ; iitr != iend; ++iitr) {
				printf("(%d,%d)", iitr->x, iitr->y);
				printf(" %dx%d\n", iitr->x_size, iitr->y_size);
			}
		}
	}
	
	void out(Bitmap *b, const std::string& name, const std::string& path)
	{
		std::string asm_s;
		std::string asm_h;
		std::string enum_h;
		std::string ptn_table;
		std::string ptn_c;
		std::string ptn_h;
		std::string ref_c;
		std::string ref_h;
		char buf[256];

		std::vector<Stock>::iterator citr = stockArray.begin();
		std::vector<Stock>::iterator cend = stockArray.end();

		// Chr Assembly
		sprintf(buf, "\t@ %s\n",citr->obj_name.c_str());
		asm_s  = "\t.section\t.rodata\n";
		asm_s += "\t.align\t2\n";
		sprintf(buf, "\t.global\t%s_chr\n\t.type\t%s_chr, object\n%s_chr:\n", name.c_str(), name.c_str(), name.c_str());
		asm_s += buf;

		// Chr Header
		sprintf(buf, "#ifndef %s_CHR_H\n#define %s_CHR_H\n#include <gba.h>\ntypedef struct { u16 chr[64]; } CHR_PAT;\nextern const CHR_PAT %s_chr[];\n#endif\n", 
			name.c_str(),name.c_str(),name.c_str());
		asm_h = buf;

		// Enum
		sprintf(buf, "#ifndef %s_ENUM_H\n#define %s_ENUM_H\nenum {\n", 
			name.c_str(),name.c_str());
		enum_h = buf;

		// Pattern
		sprintf(buf, "#include \"%s_pat.h\"\n\n", name.c_str());
		ptn_c = buf;
		
		// Pattern Pointer Table
		sprintf(buf, "const SPR_PATTERN *%s_pattern_table[] = {\n", name.c_str());
		ptn_table = buf;

		// Pattern Header
		sprintf(buf, "#ifndef %s_PAT_H\n#define %s_PAT_H\n", name.c_str(),name.c_str());
		ptn_h = buf;
		sprintf(buf, "#include <gba.h>\n#include <sprite16.h>\n#include \"%s_enum.h\"\n#include \"%s.h\"\nextern const SPR_PATTERN *%s_pattern_table[];\n", name.c_str(), name.c_str(), name.c_str());
		ptn_h += buf;


		for ( ; citr != cend; ++citr) {
			if (!b->isLoaded(citr->bmp_name)) {
				fprintf(stderr, "[%d] not found. obj:%d skipped...\n", citr->bmp_name.c_str(), citr->obj_name.c_str());
				continue;
			}

			// Dxt̂߁AЂԂ
			std::reverse(citr->info.begin(), citr->info.end());

			std::vector<Info>::iterator iitr = citr->info.begin();
			std::vector<Info>::iterator iend = citr->info.end();

//			printf("%s : ",citr->obj_name.c_str());
			const char* hvflip_str[] = {
				"SP_NOFLIP",
				"SP_HFLIP",
				"SP_VFLIP",
				"SP_HVFLIP",
			};

			int sz = 0;
			std::string ptn_c_temp;

			for ( ; iitr != iend; ++iitr) {
				int xx_ofs = 0;
				int yy_ofs = 0;
				int xx_sign = 1;
				int yy_sign = 1;
				if (iitr->hvflip & 1) { // HFLIP
					xx_ofs = -16;
					xx_sign = -1;
				}
				if (iitr->hvflip & 2) { // VFLIP
					yy_ofs = -16;
					yy_sign = -1;
				}

				sprintf(buf, "\t%s,\n",citr->obj_name.c_str());
				enum_h += buf;

				for (int yy = 0; yy < iitr->y_size / 16; ++yy) {
					for (int xx = 0; xx < iitr->x_size / 16; ++xx) {
						int hvflip = iitr->hvflip & 3;
						int index = b->generate16x16( citr->bmp_name.c_str(), 
													  iitr->x + xx * 16, 
													  iitr->y + yy * 16, 
													  &hvflip
													  );
						if (index < 0) continue;

						sz++;
						sprintf(buf, "\t\t{ %4d, %3d, %3d, %s, },\n", 
//									name.c_str(), 
									index, 
									xx_sign * xx * 16 - iitr->x_rel * xx_sign + xx_ofs, 
									yy_sign * yy * 16 - iitr->y_rel * yy_sign + yy_ofs, 
									hvflip_str[hvflip]
								);
						ptn_c_temp += buf;

					}
				}
			}
			if (sz) {
				sprintf(buf, "const SPR_PATTERN %s_ptn = {\n\t\t%d,\n\t{\n", 
							citr->obj_name.c_str(), 
							sz
						);
				ptn_c += buf;
				ptn_c += ptn_c_temp;
				ptn_c += "\t},\n};\n";

				sprintf(buf, "\t[%s] = &%s_ptn ,\n", 
							citr->obj_name.c_str(), 
							citr->obj_name.c_str()
						);
				ptn_table += buf;
			}
		}

		int stored_sprite_nums = b->getIndexSize();
		for (int i = 0; i < stored_sprite_nums; i++) {
			unsigned short *p = b->getIndexedSprPointer(i);
			for (int y = 0; y < 4; y++) {
				asm_s += "\t.short ";
				for (int x = 0; x < 16; x++) {
					sprintf(buf, "0x%04X", *p++);
					asm_s += buf;
					if (x != 15) 
						asm_s += ",";
				}
				asm_s += "\n";
			}
		}
		asm_s += "\t.ident \"bmp2s\"\n";
		enum_h += "};\n#endif\n";
		ptn_table += "};\n";
		ptn_c += ptn_table;
		ptn_h += "#endif\n";

		FILE *out;

		// Chr Assembly
		sprintf(buf, "%s.s", path.c_str());
		out = fopen(buf, "w");
		fprintf(out, "%s", asm_s.c_str());
		fclose(out);

		// Chr Header
		sprintf(buf, "%s.h", path.c_str());
		out = fopen(buf, "w");
		fprintf(out, "%s", asm_h.c_str());
		fclose(out);

		// Enum
		sprintf(buf, "%s_enum.h", path.c_str());
		out = fopen(buf, "w");
		fprintf(out, "%s", enum_h.c_str());
		fclose(out);

		// Pattern Source
		sprintf(buf, "%s_pat.c", path.c_str());
		out = fopen(buf, "w");
		fprintf(out, "%s", ptn_c.c_str());
		fclose(out);

		// Pattern Header
		sprintf(buf, "%s_pat.h", path.c_str());
		out = fopen(buf, "w");
		fprintf(out, "%s", ptn_h.c_str());
		fclose(out);
	}

};

//=============================================================================
bool minus = false;
CStock g_stk;
enum error_return {
	NO_ERR,
	FILE_OPEN_ERROR,
} error = NO_ERR;

//=============================================================================
void 
push_num(const int i)
{
	int num = i;
	if (minus) {
		num *= -1;
		minus = false;
	}
	push(num);
}

void 
defMinus(const char a)
{
	minus = true;
}

void 
defFilename(const char* const a, const char* const b)
{
	int len = b - a;
	std::string str(a, len);
	g_stk.now_bmp = str;
	if (!bmp.load(str))
	{
		error = FILE_OPEN_ERROR;
	}
}
void 
defObjname(const char* const a, const char* const b)
{
	int len = b - a;
	std::string str(a, len);
	g_stk.now_obj = str;
}
void 
defXpos(const char* const a, const char* const b)
{
	int i = pop();
	g_stk.tmp_info.x = i;
}
void 
defYpos(const char* const a, const char* const b)
{
	int i = pop();
	g_stk.tmp_info.y = i;
}
void 
defXrel(const char* const a, const char* const b)
{
	int i = pop();
	g_stk.tmp_info.x_rel = i;
}
void 
defYrel(const char* const a, const char* const b)
{
	int i = pop();
	g_stk.tmp_info.y_rel = i;
}
void 
defSize(const char* const a, const char* const b)
{
	int y = pop();
	int x = pop();
	
	if (x & 0xf)
		x = (x & ~0xf) + 0x10 * ((x < 0) ? -1 : 1);
	if (y & 0xf)
		y = (y & ~0xf) + 0x10 * ((y < 0) ? -1 : 1);

	g_stk.tmp_info.x_size = x;
	g_stk.tmp_info.y_size = y;
}

void defHVflip(const char* const a, const char* const b)
{
	CStock::Info &t = g_stk.tmp_info;
	char* c = const_cast<char *>(a);
	while (c != b) {
		if (*c == 'n') t.hvflip = 0;
		if (*c == 'N') t.hvflip = 0;
		if (*c == 'h') t.hvflip |= 1;
		if (*c == 'H') t.hvflip |= 1;
		if (*c == 'v') t.hvflip |= 2;
		if (*c == 'V') t.hvflip |= 2;
		c++;
	}
}

void 
defStock(const char* const a, const char* const b)
{
	int xsz;
	int ysz;

	CStock::Info& t = g_stk.tmp_info;
	CStock::Info& t_old = g_stk.tmp_info_old;
	if (g_stk.now_obj == "") {
		g_stk.initTemp();
		return;
	}
	if (t.y_size == -1) {
		t.x_size = t_old.x_size;
		t.y_size = t_old.y_size;
	}
	xsz = t_old.x_size;
	ysz = t_old.y_size;
	if (t.y == -1) {
		t.x = t_old.x;
		t.y = t_old.y;
		int h = bmp.getHeight(g_stk.now_bmp);
		int w = bmp.getWidth(g_stk.now_bmp);
		if (t.x + xsz + t.x_size > w) {
			t.x = 0;
			t.y += ysz;
		} else {
			t.x += xsz;
		}
	}

	t.x_rel += t.x_size / 2;
	t.y_rel += t.y_size / 2;

#if defined(DEBUG)
	printf("obj:%16s %16s\t%d,%d\t%dx%d\t[%d,%d]\t%d\n", 
		g_stk.now_bmp.c_str(),
		g_stk.now_obj.c_str(), 
		t.x, 
		t.y, 
		t.x_size, 
		t.y_size,
		t.x_rel,
		t.y_rel,
		t.hvflip
		);
#endif

	CStock::Stock &s = g_stk.tmp_stock;
	if (g_stk.now_obj == "+") {
		s.info.push_back(t);
	} else
	if (s.info.size()) {
		g_stk.stockArray.push_back(s);
		s.info.clear();
		s.bmp_name = g_stk.now_bmp;
		s.obj_name = g_stk.now_obj;
		s.info.push_back(t);
	} else {
		s.bmp_name = g_stk.now_bmp;
		s.obj_name = g_stk.now_obj;
		s.info.push_back(t);
	}
	t_old = t;
	g_stk.initTemp();
}

void 
addLastStock(const char* const a, const char* const b)
{
	CStock::Info& t = g_stk.tmp_info;
	CStock::Info& t_old = g_stk.tmp_info_old;
	CStock::Stock &s = g_stk.tmp_stock;
	if (g_stk.now_obj == "+") {
		s.info.push_back(t);
	} else
	if (s.info.size()) {
		g_stk.stockArray.push_back(s);
	} else {
		s.bmp_name = g_stk.now_bmp;
		s.obj_name = g_stk.now_obj;
		s.info.push_back(t);
		g_stk.stockArray.push_back(s);
	}
	t_old = t;
	g_stk.initTemp();
}

//=============================================================================
// 
//=============================================================================
int main(int argc, char **argv)
{
	std::string str;
	char strbuf[512];
	int index;
	for (index = 1; index < argc; index++) {
		if (*argv[index] == '-') 
			parse_opt(argv[index] + 1);
		else 
			break;
	}
	if (index == argc) {
		fprintf(stderr, __USAGE__);
		return 1;
	}

	FILE* fp;
	if ((fp = fopen(argv[index], "r")) == NULL) {
		fprintf(stderr, "Error: Can't open file '%s'\n", argv[index]);
		return 1;
	}
	while(NULL != fgets(strbuf, 512, fp))
	{
		str += strbuf;
	}
	fclose(fp);

	if(!parse_txt(str))
		fprintf(stderr, "Illegal input file\n");

	std::string basename; // only filename
	std::string pathname; // with path
	if (argc > index+1) {
		pathname = argv[index+1];
		get_basename(basename, argv[index+1]);
	} else {
		get_basename(basename, argv[index]);
		pathname = basename;
	}

	if(debug_print) {
		g_stk.print(&bmp);
	} else 
	if (error == NO_ERR) {
		g_stk.out(&bmp, basename, pathname);
	}

	return error;
}

//
void get_basename(std::string &str, const char *c)
{
	char tmp[256];
	char *ps;
	char *pe;
	size_t len;

	ps = strrchr(c, '/');
	if (ps == NULL) 
		ps = strrchr(c, '\\');

	if (ps == NULL) 
		ps = const_cast<char *>(c);
	else 
		ps++;

	pe = strrchr(c, '.');
	if (pe == NULL || pe < ps)
		pe = strrchr(c, 0);

	len = pe - ps;

	strncpy(tmp, ps, pe-ps);
	tmp[len] = '\0';
	str = tmp;
}

//-----------------------------------------------------------------------------
// parse option
//-----------------------------------------------------------------------------
void parse_opt(char *s)
{
	if (*s != '\0') {
		switch (*s) {
		case 'd':
			debug_print = true;
			break;
		default:
			break;
		}
	}
	return;
}
