#include "act.hpp"
#include <boost/spirit/core.hpp>
#include <boost/spirit/utility/confix.hpp>
using namespace boost::spirit;

struct comment_parser : public grammar<comment_parser> {
	template<typename S> struct definition {
		rule<S> comment;
		definition(const comment_parser& self) {
			comment =
				comment_p("//")
				| comment_p("/*", "*/")
				;
		}
		const rule<S>& start() const { return comment; }
	};
};

struct txt_parser : public grammar<txt_parser> 
{
	template<typename S> 
	struct definition 
	{
		rule<S> txt, def, filename, objname, position, size, num_p, rel_pos, hvflip;

		definition(const txt_parser& self)
		{
			num_p = int_p[&push_num] | (!ch_p('-')[&defMinus] >> ch_p('$') >> hex_p[&push_num]);

			filename = '*'
						>> (+(anychar_p - '*' - space_p - '[' - ']'))[&defFilename]
						>> +eol_p >> *space_p;

			objname  = *blank_p >> (ch_p('+') | ((anychar_p - '*' - '+' - '[' - ']' - space_p - int_p) >> +(anychar_p - '*' - space_p)))[&defObjname]
						>> *blank_p;

			position = (num_p[&defXpos]
						>> *blank_p 
						>> ch_p(',') 
						>> *blank_p 
						>> num_p[&defYpos])
						>> *blank_p;

			size     =  (num_p >> ch_p('x') >> num_p)[&defSize] >> *blank_p;

			rel_pos  = ch_p('[') >> *blank_p >> num_p[&defXrel] >> *blank_p >> ch_p(',') >> *blank_p >> num_p[&defYrel] >> *blank_p >> ch_p(']') >> *blank_p;

			hvflip   = (+(ch_p('N') | ch_p('H') | ch_p('V') | ch_p('n') | ch_p('h') | ch_p('v')))[&defHVflip] >> *blank_p;

			def = filename >> *( objname >> !position >> !size >> !rel_pos >> !hvflip >> eol_p >> eps_p[&defStock] >> *space_p) >> *space_p;
			txt = *space_p >> +def >> eps_p[&addLastStock];
		}

		const rule<S>& start() const { return txt; }
	};
};

bool parse_txt(std::string str) {
	txt_parser txt_parse;
	comment_parser comment;
	parse_info<> r = parse(str.c_str(), txt_parse, comment);
	
	if (r.hit)
		return true;
	else
		return false;
}
